"""cosmian_lib.sgx.crypto_lib module."""

from ctypes import *
from pathlib import Path
from typing import Tuple


sgx_crypto_lib_path = Path("/usr/local/lib/libcs_sgx_crypto.so")

if not sgx_crypto_lib_path.exists():
    raise FileNotFoundError(f"Can't find '{sgx_crypto_lib_path}'")

crypto_lib = cdll.LoadLibrary(sgx_crypto_lib_path)

# int cs_sgx__init()
crypto_lib.cs_sgx__init.argtypes = []
crypto_lib.cs_sgx__init.restype = c_int

# cs_sgx__running_inside_enclave()
crypto_lib.cs_sgx__running_inside_enclave.argtypes = []
crypto_lib.cs_sgx__running_inside_enclave.restype = c_int

# int cs_sgx__encrypt_code(unsigned char *encrypted_code, const unsigned char *code, unsigned long long code_length,
#                         const unsigned char *symkey);
crypto_lib.cs_sgx__encrypt_code.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte), c_ulonglong, POINTER(c_ubyte)]
crypto_lib.cs_sgx__encrypt_code.restype = c_int

# int cs_sgx__encrypt_code_in_enclave(unsigned char *encrypted_code, const unsigned char *code,
#                                    unsigned long long code_length);
crypto_lib.cs_sgx__encrypt_code_in_enclave.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte), c_ulonglong]
crypto_lib.cs_sgx__encrypt_code_in_enclave.restype = c_int

# int cs_sgx__decrypt_code(unsigned char *code, const unsigned char *encrypted_code,
#                         unsigned long long encrypted_code_length, const unsigned char *symkey);
crypto_lib.cs_sgx__decrypt_code.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte), c_ulonglong, POINTER(c_ubyte)]
crypto_lib.cs_sgx__decrypt_code.restype = c_int

# int cs_sgx__decrypt_code_in_enclave(unsigned char *code, const unsigned char *encrypted_code,
#                                     unsigned long long encrypted_code_length);
crypto_lib.cs_sgx__decrypt_code_in_enclave.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte), c_ulonglong]
crypto_lib.cs_sgx__decrypt_code_in_enclave.restype = c_int

# int cs_sgx__box_symkey(unsigned char *boxed_symkey, const unsigned char *algo_provider_x25519_pk);
crypto_lib.cs_sgx__box_symkey.argtypes = [POINTER(c_ubyte), POINTER(c_ubyte)]
crypto_lib.cs_sgx__box_symkey.restype = c_int

# int cs_sgx__enclave_x25519_key_pair(unsigned char *pk, unsigned char *sk);
crypto_lib.cs_sgx__enclave_x25519_key_pair.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte)]
crypto_lib.cs_sgx__enclave_x25519_key_pair.restype = c_int

# int cs_sgx__hash_shared_key(unsigned char *shared_key, const unsigned char *scalar_mult,
#                             const unsigned char *client_pk, const unsigned char *server_pk)
crypto_lib.cs_sgx__hash_shared_key.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte), POINTER(c_ubyte), POINTER(c_ubyte)
]
crypto_lib.cs_sgx__hash_shared_key.restype = c_int

# int cs_sgx__generate_x25519_shared_key(unsigned char *shared_key, const unsigned char *sk,
#                                        const unsigned char *remote_pk)
crypto_lib.cs_sgx__generate_x25519_shared_key.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte), POINTER(c_ubyte)
]
crypto_lib.cs_sgx__generate_x25519_shared_key.restype = c_int

# int cs_sgx__generate_exponentiation(unsigned char *pk, unsigned char *sk);
crypto_lib.cs_sgx__generate_exponentiation.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte)]
crypto_lib.cs_sgx__generate_exponentiation.restype = c_int

# int cs_sgx__generate_exponentiation(unsigned char *pk, unsigned char *sk);
crypto_lib.cs_sgx__generate_data_provider_shared_key.argtypes = [
    POINTER(c_ubyte), POINTER(c_ubyte), POINTER(c_ubyte), POINTER(c_ubyte)]
crypto_lib.cs_sgx__generate_data_provider_shared_key.restype = c_int

# int cs_sgx__get_quote(const unsigned char* data_provider_pk, char* b64_quote);
crypto_lib.cs_sgx__get_quote.argtypes = [
    POINTER(c_ubyte), c_char_p]
crypto_lib.cs_sgx__get_quote.restype = c_int

# getkey_wrapper()s
crypto_lib.sgx__cs__run_all_tests.argtypes = []
crypto_lib.sgx__cs__run_all_tests.restype = c_int

if crypto_lib.cs_sgx__init() != 0:
    raise ValueError('failed initializing the crypto library')


class SGXCrypto:

    @staticmethod
    def init() -> bool:
        return crypto_lib.cs_sgx__init() == 0

    @staticmethod
    def is_running_in_enclave() -> bool:
        return crypto_lib.cs_sgx__running_inside_enclave() == 0

    @staticmethod
    def encrypt_code(code: bytes, key: bytes, sgx: bool = False) -> bytes:
        code_length = len(code)
        code_bytes_c_array = (c_ubyte * code_length)(*code)
        key_length = len(key)
        key_c_array = (c_ubyte * len(key))(*key)
        encrypted_code = (c_ubyte * (code_length + 40))()

        if not sgx:
            res = crypto_lib.cs_sgx__encrypt_code(
                    cast(encrypted_code, POINTER(c_ubyte)),
                    code_bytes_c_array,
                    code_length,
                    key_c_array
            )
        else:
            if not SGXCrypto.is_running_in_enclave():
                raise ValueError(
                    "You're code is not running inside SGX enclave!"
                )
            # print("decryption of shared key inside the enclave")
            dec_sharedkey = (c_ubyte * (key_length - 40))()
            res = crypto_lib.cs_sgx__decrypt_code_in_enclave(
                cast(dec_sharedkey, POINTER(c_ubyte)),
                key_c_array,
                key_length
            )
            if res != 0:
                raise ValueError("decryption of shared key failed")

            # print("decryption of code inside the enclave")
            dec_sharedkey_c_array = (
                c_ubyte * len(dec_sharedkey))(*dec_sharedkey)
            res = crypto_lib.cs_sgx__encrypt_code(
                cast(encrypted_code, POINTER(c_ubyte)),
                code_bytes_c_array,
                code_length,
                dec_sharedkey_c_array
            )

        if res != 0:
            raise ValueError("encryption of code failed")
        return bytes(encrypted_code)

    @staticmethod
    def decrypt_code(encrypted_code: bytes, key: bytes, sgx: bool = False) -> bytes:
        encrypted_code_c_array = (
            c_ubyte * len(encrypted_code))(*encrypted_code)
        encrypted_code_length = len(encrypted_code)
        key_length = len(key)
        key_c_array = (c_ubyte * len(key))(*key)
        code = (c_ubyte * (encrypted_code_length - 40))()

        if not sgx:
            # use key as it is
            res = crypto_lib.cs_sgx__decrypt_code(
                cast(code, POINTER(c_ubyte)),
                encrypted_code_c_array,
                encrypted_code_length,
                key_c_array
            )
        else:
            if not SGXCrypto.is_running_in_enclave():
                raise ValueError(
                    "You're code is not running inside SGX enclave!"
                )
            # print("decryption of shared key inside the enclave")
            dec_sharedkey = (c_ubyte * (key_length - 40))()
            res = crypto_lib.cs_sgx__decrypt_code_in_enclave(
                cast(dec_sharedkey, POINTER(c_ubyte)),
                key_c_array,
                key_length
            )
            if res != 0:
                raise ValueError("decryption of shared key failed")

            # print("decryption of code inside the enclave")
            dec_sharedkey_c_array = (
                c_ubyte * len(dec_sharedkey))(*dec_sharedkey)
            res = crypto_lib.cs_sgx__decrypt_code(
                cast(code, POINTER(c_ubyte)),
                encrypted_code_c_array,
                encrypted_code_length,
                dec_sharedkey_c_array
            )

        if res != 0:
            raise ValueError("decryption of code failed")
        return bytes(code)

    @staticmethod
    def box_symkey(algo_provider_x25519_pk: bytes) -> bytes:
        # this method can only be run inside an enclave
        if crypto_lib.cs_sgx__running_inside_enclave() != 0:
            raise ValueError(
                'boxing of a symkey can only happen in a Graphene enclave')
        #
        boxed_symkey = (c_ubyte * 72)()
        algo_provider_x25519_pk_c_array = (
            c_ubyte * len(algo_provider_x25519_pk))(*algo_provider_x25519_pk)
        if crypto_lib.cs_sgx__box_symkey(
                cast(boxed_symkey, POINTER(c_ubyte)),
                algo_provider_x25519_pk_c_array
        ) != 0:
            raise ValueError("boxing of the symkey failed")
        return bytes(boxed_symkey)

    @staticmethod
    def enclave_x25519_key_pair() -> Tuple[bytes, bytes]:
        # this method can only be run inside an enclave
        if crypto_lib.cs_sgx__running_inside_enclave() != 0:
            raise ValueError(
                'the enclave x25519 key pair can only be retrieved inside SGX enclave')
        #
        pk = (c_ubyte * 32)()
        sk = (c_ubyte * 32)()
        if crypto_lib.cs_sgx__enclave_x25519_key_pair(
                cast(pk, POINTER(c_ubyte)),
                cast(sk, POINTER(c_ubyte))
        ) != 0:
            raise ValueError("retrieving the enclave x25519 key pair failed")
        return bytes(pk), bytes(sk)

    @staticmethod
    def enclave_exponentiation_key_pair() -> Tuple[bytes, bytes]:
        # this method can only be run inside an enclave
        if crypto_lib.cs_sgx__running_inside_enclave() != 0:
            raise ValueError(
                'the enclave exponentiation key pair cannot only be retrieved from inside a Graphene enclave')
        #
        pk = (c_ubyte * 32)()
        sk = (c_ubyte * 32)()
        if crypto_lib.cs_sgx__generate_exponentiation(
                cast(pk, POINTER(c_ubyte)),
                cast(sk, POINTER(c_ubyte))
        ) != 0:
            raise ValueError(
                "retrieving the enclave exponentiation key pair failed")
        return bytes(pk), bytes(sk)

    @staticmethod
    def enclave_shared_key(pk: bytes, sk: bytes, algo_provider_pk: bytes) -> bytes:
        # this method can only be run inside an enclave
        if crypto_lib.cs_sgx__running_inside_enclave() != 0:
            raise ValueError(
                'the enclave shared key cannot only be retrieved from inside a Graphene enclave')
        #
        shared_key = (c_ubyte * 32)()
        pk_c_array = (c_ubyte * len(pk))(*pk)
        sk_c_array = (c_ubyte * len(sk))(*sk)
        algo_provider_pk_c_array = (
            c_ubyte * len(algo_provider_pk))(*algo_provider_pk)
        if crypto_lib.cs_sgx__generate_data_provider_shared_key(
                cast(shared_key, POINTER(c_ubyte)),
                pk_c_array,
                sk_c_array,
                algo_provider_pk_c_array
        ) != 0:
            raise ValueError(
                "retrieving the enclave shared key failed")
        return bytes(shared_key)

    @staticmethod
    def enclave_x25519(public_key: bytes) -> Tuple[bytes, bytes]:
        if crypto_lib.cs_sgx__running_inside_enclave() != 0:
            raise ValueError('shared key can only be retrieved inside SGX enclave')

        enclave_pubkey, enclave_privkey = SGXCrypto.enclave_x25519_key_pair()

        shared_key = (c_ubyte * 32)()
        shared_key_hash = (c_ubyte * 32)()

        enclave_pubkey_c_array = (c_ubyte * len(enclave_pubkey))(*enclave_pubkey)
        enclave_privkey_c_array = (c_ubyte * len(enclave_privkey))(*enclave_privkey)
        public_key_c_array = (c_ubyte * len(public_key))(*public_key)

        # compute enclave_privkey * public_key
        if crypto_lib.cs_sgx__generate_x25519_shared_key(
            cast(shared_key, POINTER(c_ubyte)),
            enclave_privkey_c_array,
            public_key_c_array
        ) != 0:
            raise ValueError("compute enclave shared key failed")

        # compute H(shared_key || public_key || enclave_pubkey)
        if crypto_lib.cs_sgx__hash_shared_key(
            cast(shared_key_hash, POINTER(c_ubyte)),
            cast(shared_key, POINTER(c_ubyte)),
            public_key_c_array,
            enclave_pubkey_c_array
        ) != 0:
            raise ValueError("hash enclave shared key failed")

        return enclave_pubkey, bytes(shared_key_hash)

    @staticmethod
    def enclave_encrypt_shared_key(shared_key: bytes) -> bytes:
        # this method can only be run inside an enclave
        if crypto_lib.cs_sgx__running_inside_enclave() != 0:
            raise ValueError(
                'the encrypted shared key cannot only be retrieved from inside a Graphene enclave')
        #
        shared_key_len = len(shared_key)
        shared_key_c_array = (c_ubyte * shared_key_len)(*shared_key)
        encrypted_shared_key = (c_ubyte * (shared_key_len + 40))()
        if crypto_lib.cs_sgx__encrypt_code_in_enclave(
                cast(encrypted_shared_key, POINTER(c_ubyte)),
                shared_key_c_array,
                shared_key_len
        ) != 0:
            raise ValueError('encryption of shared key failed')
        return bytes(encrypted_shared_key)

    @staticmethod
    def enclave_decrypt_shared_key(enc_shared_key: bytes) -> bytes:
        # this method can only be run inside an enclave
        if crypto_lib.cs_sgx__running_inside_enclave() != 0:
            raise ValueError(
                'the decrypted shared key cannot only be retrieved from inside a Graphene enclave')
        #
        enc_shared_key_len = len(enc_shared_key)
        enc_shared_key_c_array = (
            c_ubyte * enc_shared_key_len)(*enc_shared_key)
        decrypted_shared_key = (c_ubyte * (enc_shared_key_len - 40))()
        if crypto_lib.cs_sgx__decrypt_code_in_enclave(
                cast(decrypted_shared_key, POINTER(c_ubyte)),
                enc_shared_key_c_array,
                enc_shared_key_len
        ) != 0:
            raise ValueError('decryption of shared key failed')
        return bytes(decrypted_shared_key)

    @staticmethod
    def enclave_get_quote(data_provider_pk: bytes) -> str:
        # this method can only be run inside an enclave
        if crypto_lib.cs_sgx__running_inside_enclave() != 0:
            raise ValueError(
                'the get quote cannot only be retrieved from inside a Graphene enclave')
        #
        quote = (c_char * 8192)()
        data_provider_pk_c_array = (
            c_ubyte * len(data_provider_pk))(*data_provider_pk)
        if crypto_lib.cs_sgx__get_quote(
                data_provider_pk_c_array,
                cast(quote, c_char_p)
        ) != 0:
            raise ValueError('retrieving the quote failed')
        return bytes(quote).decode('utf-8').rstrip('\0')
